#include <GenTL.h>
#include <chrono>
#include <exception>
#include <iostream>
#include <string>
#include <vector>
#include "Zivid/GenTLAddresses.h"

int main()
{
    try
    {
        auto errorCode = GenTL::GCInitLib();
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed initializing GenTL library. Error code: " + std::to_string(errorCode) };

        GenTL::TL_HANDLE tlHandle{};
        errorCode = GenTL::TLOpen(&tlHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed initializing transport layer. Error code: " + std::to_string(errorCode) };

        bool8_t changedInterfaceList{ false };
        uint64_t timeout{ 10 };

        errorCode = GenTL::TLUpdateInterfaceList(tlHandle, &changedInterfaceList, timeout);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed updating interface list. Error code: " + std::to_string(errorCode) };
        if(!changedInterfaceList) throw std::runtime_error{ "Failed finding any interfaces." };

        uint32_t numInterfaces{ 0 };
        errorCode = GenTL::TLGetNumInterfaces(tlHandle, &numInterfaces);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed getting number of interfaces. Error code: " + std::to_string(errorCode) };
        if(numInterfaces <= 0) throw std::runtime_error{ "Failed to find any interfaces." };

        char idBuffer[100];
        size_t idBufferSize{ sizeof(idBuffer) };

        errorCode = GenTL::TLGetInterfaceID(tlHandle, 0, idBuffer, &idBufferSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed getting interface IDs. Error code: " + std::to_string(errorCode) };

        GenTL::IF_HANDLE ifHandle;
        errorCode = GenTL::TLOpenInterface(tlHandle, idBuffer, &ifHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed opening interface. Error code: " + std::to_string(errorCode) };

        uint64_t deviceTimeout{ 100 };
        bool8_t changedDeviceList{ false };
        errorCode = GenTL::IFUpdateDeviceList(ifHandle, &changedDeviceList, deviceTimeout);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed updating device list. Error code: " + std::to_string(errorCode) };
        if(!changedDeviceList)
            throw std::runtime_error{ "Failed finding any devices. Changed device list: "
                                      + std::to_string(changedDeviceList) };

        uint32_t numDevices{ 0 };
        errorCode = GenTL::IFGetNumDevices(ifHandle, &numDevices);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed getting number of devices. Error code: " + std::to_string(errorCode) };
        if(numDevices <= 0)
            throw std::runtime_error{ "No camera found. Number of devices: " + std::to_string(numDevices) };

        GenTL::DEVICE_ACCESS_STATUS deviceAccessFlags{ GenTL::DEVICE_ACCESS_EXCLUSIVE };
        char deviceIDBuffer[100];
        size_t deviceIDBufferSize{ sizeof(deviceIDBuffer) };

        errorCode = GenTL::IFGetDeviceID(ifHandle, 0, deviceIDBuffer, &deviceIDBufferSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed getting device ID. Error code: " + std::to_string(errorCode) };

        std::cout << "Connecting to camera" << std::endl;
        GenTL::DEV_HANDLE devHandle;
        errorCode = GenTL::IFOpenDevice(ifHandle, deviceIDBuffer, deviceAccessFlags, &devHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed opening device. Error code: " + std::to_string(errorCode) };

        GenTL::PORT_HANDLE remDevHandle;
        errorCode = GenTL::DevGetPort(devHandle, &remDevHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed opening remote device. Error code: " + std::to_string(errorCode) };

        std::cout << "Adjusting the camera settings" << std::endl;
        size_t iris{ 20 };
        size_t irisSize{ sizeof(iris) };
        errorCode =
            GenTL::GCWritePort(remDevHandle, ZividGenTL::RegisterAddresses::AnalogControl::iris, &iris, &irisSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to set iris. Error code: " + std::to_string(errorCode) };

        double exposureTime{ 8333 };
        size_t exposureTimeSize{ sizeof(exposureTime) };

        errorCode = GenTL::GCWritePort(remDevHandle,
                                       ZividGenTL::RegisterAddresses::AcquisitionControl::exposureTime,
                                       &exposureTime,
                                       &exposureTimeSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to set exposure time. Error code: " + std::to_string(errorCode) };
        bool8_t outlierEnabled{ true };
        size_t outlierEnabledSize{ sizeof(outlierEnabled) };

        errorCode = GenTL::GCWritePort(remDevHandle,
                                       ZividGenTL::RegisterAddresses::AcquisitionControl::outlierFilterEnabled,
                                       &outlierEnabled,
                                       &outlierEnabledSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to enabled outlier filter. Error code: " + std::to_string(errorCode) };
        double outlierThreshold{ 5 };
        size_t outlierThresholdSize{ sizeof(outlierThreshold) };

        errorCode = GenTL::GCWritePort(remDevHandle,
                                       ZividGenTL::RegisterAddresses::AcquisitionControl::outlierFilterThreshold,
                                       &outlierThreshold,
                                       &outlierThresholdSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to set outlier filter threshold. Error code: "
                                      + std::to_string(errorCode) };

        uint32_t numStreams{ 0 };
        errorCode = GenTL::DevGetNumDataStreams(devHandle, &numStreams);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed getting number of datastreams. Error code: "
                                      + std::to_string(errorCode) };
        if(numStreams <= 0)
            throw std::runtime_error{ "No datastreams found. Number of streams: " + std::to_string(numStreams) };

        char dataStreamIdBuffer[1024];
        size_t dataStreamBufferSize = sizeof(dataStreamIdBuffer);
        errorCode = GenTL::DevGetDataStreamID(devHandle, 0, dataStreamIdBuffer, &dataStreamBufferSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed getting datastream ID. Error code: " + std::to_string(errorCode) };

        GenTL::DS_HANDLE dsHandle;
        errorCode = GenTL::DevOpenDataStream(devHandle, dataStreamIdBuffer, &dsHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed opening datastream. Error code: " + std::to_string(errorCode) };

        GenTL::BUFFER_HANDLE bufferHandle{ GENTL_INVALID_HANDLE };
        char privateBufferData[100];

        errorCode = GenTL::DSAllocAndAnnounceBuffer(dsHandle, 46080000, privateBufferData, &bufferHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed announcing and allocating buffers. Error code: "
                                      + std::to_string(errorCode) };

        errorCode = GenTL::DSQueueBuffer(dsHandle, bufferHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to queue buffer. Error code: " + std::to_string(errorCode) };

        GenTL::EVENT_HANDLE newBufferEvent{ GENTL_INVALID_HANDLE };
        errorCode = GCRegisterEvent(dsHandle, GenTL::EVENT_NEW_BUFFER, &newBufferEvent);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to register event. Error code: " + std::to_string(errorCode) };

        uint64_t numImages{ 1 };
        uint32_t acqusitionMode{ 0 }; // SingleCapture
        size_t acquisitionModeBufferSize = sizeof(acqusitionMode);

        errorCode = GenTL::GCWritePort(remDevHandle,
                                       ZividGenTL::RegisterAddresses::AcquisitionControl::acquisitionMode,
                                       &acqusitionMode,
                                       &acquisitionModeBufferSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to change to SingleCapture. Error code: " + std::to_string(errorCode) };

        errorCode = DSStartAcquisition(dsHandle, GenTL::ACQ_START_FLAGS_DEFAULT, numImages);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to start acquisition loop. Error code: " + std::to_string(errorCode) };

        bool acquisitionStart{ true };
        size_t acquisitionStartSize = sizeof(acquisitionStart);
        errorCode = GenTL::GCWritePort(remDevHandle,
                                       ZividGenTL::RegisterAddresses::AcquisitionControl::acquisitionStart,
                                       &acquisitionStart,
                                       &acquisitionStartSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed set acquisition started. Error code: " + std::to_string(errorCode) };

        GenTL::EVENT_NEW_BUFFER_DATA newBufferData{};
        size_t newBufferDataSize = sizeof(newBufferData);

        std::cout << "Capture a frame" << std::endl;
        errorCode = GenTL::EventGetData(newBufferEvent, &newBufferData, &newBufferDataSize, 10000);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to capture. Error code: " + std::to_string(errorCode) };

        bool acquisitonStop{ true };
        size_t acquisitonStopSize = sizeof(acquisitonStop);
        errorCode = GenTL::GCWritePort(remDevHandle,
                                       ZividGenTL::RegisterAddresses::AcquisitionControl::acquisitionStop,
                                       &acquisitonStop,
                                       &acquisitonStopSize);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed set acquisition stop. Error code: " + std::to_string(errorCode) };

        errorCode = GenTL::DSStopAcquisition(dsHandle, GenTL::ACQ_STOP_FLAGS_DEFAULT);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed to stop acquisition. Error code: " + std::to_string(errorCode) };

        errorCode = GenTL::DSClose(dsHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed closing datastream. Error code: " + std::to_string(errorCode) };

        errorCode = GenTL::DevClose(devHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed closing device. Error code: " + std::to_string(errorCode) };

        errorCode = GenTL::IFClose(ifHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed closing interface. Error code: " + std::to_string(errorCode) };

        errorCode = GenTL::TLClose(tlHandle);
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed closing transport layer. Error code: " + std::to_string(errorCode) };

        errorCode = GenTL::GCCloseLib();
        if(errorCode != GenTL::GC_ERR_SUCCESS)
            throw std::runtime_error{ "Failed uninitializing GenTL library. Error code: " + std::to_string(errorCode) };
    }
    catch(const std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
}
